# -*- coding: utf-8 -*-
"""
@file
@brief Adds functionalities to a dataframe.
"""
import datetime
import pandas


class TableFormula(pandas.DataFrame):  # pylint: disable=W0223
    """
    Extends class :epkg:`pandas:DataFrame` or proposes extensions
    to existing functions using lambda functions.
    See `Extending Pandas
    <https://pandas.pydata.org/pandas-docs/
    stable/development/extending.html>`_.
    """

    @property
    def _constructor(self):
        return TableFormula

    def sort(self, function_sort, reverse=False):
        """
        Sorts rows based on the values returned by *function_sort*.

        @param      function_sort   lambda function
        @param      reverse         reverse order

        The function creates a column ``__key__`` and removes it later.
        The changes happen inplace.
        """
        if "__key__" in self.columns:
            raise ValueError(
                "__key__ cannot be used in the original dataframe.")
        self["__key__"] = self.apply(lambda row: function_sort(row), axis=1)
        self.sort_values("__key__", inplace=True, ascending=not reverse)
        self.drop("__key__", inplace=True, axis=1)

    def fgroupby(self, function_key, function_values, columns=None,
                 function_agg=None, function_weight=None):
        """
        Groups information based on columns defined by lambda functions.

        @param      function_key        defines the key
        @param      function_values     defines the values
        @param      columns             name of the columns, if None, new ones will be created
        @param      function_agg        how to aggregate the data, if None, the default is
                                        :epkg:`pandas:DataFrame:sum`.
        @param      function_weight     defines weights, can be None

        The function uses columns ``__key__``, ``__weight__``.
        You should not use these names.
        Others columns are created ``__value_{0}__`` and
        ``__weight_{0}__``. All of them are created and removed
        before returning the result.

        Example:

        ::

            group = table.groupby(lambda v: v["name"],
                      [lambda v: v["d_a"]],
                      ["sum_d_a"],
                      [lambda vec, w: sum(vec) / w],
                      lambda v: v["d_b"])
        """
        if "__key__" in self.columns:
            raise ValueError(
                "__key__ cannot be used in the original dataframe.")
        if "__weight__" in self.columns:
            raise ValueError(
                "__weight__ cannot be used in the original dataframe.")

        cp = self.copy()
        cp["__key__"] = cp.apply(lambda row: function_key(row), axis=1)
        if function_weight is not None:
            cp["__weight__"] = cp.apply(
                lambda row: function_weight(row), axis=1)

        if columns is None:
            columns = ["fv{0}" for i in range(len(function_values))]
        if len(columns) != len(function_values):
            raise ValueError(
                "Parameters function_values and columns must have the same size.")
        if function_agg is None:
            function_agg = [pandas.DataFrame.sum for c in columns]
        if len(function_agg) != len(function_values):
            raise ValueError(
                "Parameters function_values and function_agg must have the same size.")

        values = []
        rep = dict()
        for v, cnew in zip(function_values, columns):
            n = "__value_{0}__".format(cnew)
            values.append(n)
            rep[n] = cnew
            if function_weight is None:
                cp[n] = cp.apply(lambda row, v=v: v(row), axis=1)
            else:
                cp[n] = cp.apply(lambda row, v=v: v(
                    row), axis=1) * cp["__weight__"]

        if function_weight is None:
            aggs = {k: v for k, v in zip(  # pylint: disable=R1721
                values, function_agg)}  # pylint: disable=R1721
            gr = cp.groupby("__key__", as_index=False).agg(aggs)
        else:
            sum_weight = cp["__weight__"].sum()
            aggs = {k: (lambda c, v=v: v(c, sum_weight))  # pylint: disable=W0631
                    for k, v in zip(values, function_agg)}
            gr = cp.groupby("__key__", as_index=False).agg(aggs)
        gr.columns = [rep.get(_, _) for _ in gr.columns]
        gr = gr.drop("__key__", axis=1)
        return TableFormula(gr)

    def add_column_index(self, index, name=None):
        """
        Changes the index.

        @param      index       new_index
        @param      name        name of the index

        The changes happen inplace.
        """
        self["__key__"] = index
        self.set_index("__key__", inplace=True)
        self.index.rename(name, inplace=True)

    def add_column_vector(self, name, values):
        """
        Adds a column knowing its name and a vector of values.

        @param      name                name of the column
        @param      values              values

        The changes happen inplace.
        """
        self[name] = values

    def addc(self, name, function_value):
        """
        Adds a column knowing its name and a lambda function.

        @param      name                name of the column
        @param      function_value      function

        The changes happen inplace.
        """
        self[name] = self.apply(lambda row: function_value(row), axis=1)

    def graph_XY(self, curves, xlabel=None, ylabel=None, marker=True,
                 link_point=False, title=None, format_date="%Y-%m-%d",
                 legend_loc=0, figsize=None, ax=None):
        """
        @param      curves      list of 3-uples (generator for X, generator for Y, label)
                                for some layout, it can also be:
                                (generator for X, generator for Y, generator for labels, label)
        @param      xlabel      label for X axis
        @param      ylabel      label for Y axis
        @param      marker      add a marker for each point
        @param      link_point  link points between them
        @param      title       graph title
        @param      format_date if X axis is a datetime object, the function will use this format
                                to print dates
        @param      legend_loc  location of the legend
        @param      figsize     size of the figure
        @param      ax          :epkg:`matplotlib:Axis` or None to create a new one
        @return                 :epkg:`matplotlib:Axis`

        For the legend position, see `matplotlib <http://matplotlib.org/api/legend_api.html>`_.

        Example:

        ::

            table.graph_XY ( [ [ lambda v: v["sum_a"], lambda v: v["sum_b"], "xy label 1"],
                               [ lambda v: v["sum_b"], lambda v: v["sum_c"], "xy label 2"],
                                ])
        """
        if ax is None:
            import matplotlib.pyplot as plt  # pylint: disable=C0415
            fig, ax = plt.subplots(1, 1, figsize=figsize)

        smarker = {(True, True): 'o-', (True, False): 'o', (False, True): '-',
                   # (False, False) :''
                   }[marker, link_point]

        has_date = False
        for xf, yf, label in curves:
            x = self.apply(xf, axis=1)
            y = self.apply(yf, axis=1)
            if isinstance(x[0], datetime.datetime):
                import matplotlib.dates  # pylint: disable=C0415
                x = [matplotlib.dates.date2num(d) for d in x]
                has_date = True
            ax.plot(x, y, smarker, label=label)

        if has_date:
            import matplotlib.dates  # pylint: disable=C0415
            hfmt = matplotlib.dates.DateFormatter(format_date)
            if "%H" in format_date or "%M" in format_date:
                ax.xaxis.set_major_locator(matplotlib.dates.MinuteLocator())
            ax.xaxis.set_major_formatter(hfmt)
            fig = ax.get_figure()
            fig.autofmt_xdate()

        ax.legend(loc=legend_loc)
        return ax



# Generated from C:\xadupre\github\pyensae\src\pyensae\languages\RFilter.g4 by ANTLR 4.10.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    return [
        4, 1, 70, 147, 2, 0, 7, 0, 2, 1, 7, 1, 2, 2, 7, 2, 2, 3, 7, 3, 2, 4, 7, 4, 1, 0, 1, 0, 1, 0, 5,
        0, 14, 8, 0, 10, 0, 12, 0, 17, 9, 0, 1, 0, 1, 0, 1, 1, 1, 1, 4, 1, 23, 8, 1, 11, 1, 12, 1, 24,
        1, 2, 1, 2, 3, 2, 29, 8, 2, 1, 2, 1, 2, 1, 2, 3, 2, 34, 8, 2, 1, 2, 1, 2, 1, 2, 1, 2, 5, 2, 40,
        8, 2, 10, 2, 12, 2, 43, 9, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 5, 2, 50, 8, 2, 10, 2, 12, 2, 53,
        9, 2, 1, 2, 1, 2, 1, 2, 1, 2, 5, 2, 59, 8, 2, 10, 2, 12, 2, 62, 9, 2, 1, 2, 1, 2, 1, 2, 1, 2,
        5, 2, 68, 8, 2, 10, 2, 12, 2, 71, 9, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 77, 8, 2, 1, 2, 1, 2, 1,
        2, 5, 2, 82, 8, 2, 10, 2, 12, 2, 85, 9, 2, 1, 2, 1, 2, 3, 2, 89, 8, 2, 1, 2, 1, 2, 3, 2, 93,
        8, 2, 1, 2, 1, 2, 1, 2, 5, 2, 98, 8, 2, 10, 2, 12, 2, 101, 9, 2, 1, 2, 1, 2, 3, 2, 105, 8, 2,
        1, 2, 1, 2, 3, 2, 109, 8, 2, 1, 2, 1, 2, 1, 2, 5, 2, 114, 8, 2, 10, 2, 12, 2, 117, 9, 2, 1,
        2, 1, 2, 3, 2, 121, 8, 2, 1, 2, 1, 2, 3, 2, 125, 8, 2, 1, 2, 1, 2, 1, 2, 5, 2, 130, 8, 2, 10,
        2, 12, 2, 133, 9, 2, 1, 2, 1, 2, 3, 2, 137, 8, 2, 1, 2, 1, 2, 3, 2, 141, 8, 2, 1, 3, 1, 3, 1,
        4, 1, 4, 1, 4, 0, 0, 5, 0, 2, 4, 6, 8, 0, 2, 4, 0, 15, 15, 22, 22, 25, 30, 62, 67, 12, 0, 5,
        11, 14, 14, 19, 21, 23, 23, 31, 31, 34, 34, 39, 39, 41, 41, 45, 51, 53, 53, 55, 60, 68,
        68, 182, 0, 15, 1, 0, 0, 0, 2, 22, 1, 0, 0, 0, 4, 140, 1, 0, 0, 0, 6, 142, 1, 0, 0, 0, 8, 144,
        1, 0, 0, 0, 10, 14, 3, 4, 2, 0, 11, 14, 5, 69, 0, 0, 12, 14, 5, 1, 0, 0, 13, 10, 1, 0, 0, 0,
        13, 11, 1, 0, 0, 0, 13, 12, 1, 0, 0, 0, 14, 17, 1, 0, 0, 0, 15, 13, 1, 0, 0, 0, 15, 16, 1,
        0, 0, 0, 16, 18, 1, 0, 0, 0, 17, 15, 1, 0, 0, 0, 18, 19, 5, 0, 0, 1, 19, 1, 1, 0, 0, 0, 20,
        21, 5, 69, 0, 0, 21, 23, 6, 1, -
        1, 0, 22, 20, 1, 0, 0, 0, 23, 24, 1, 0, 0, 0, 24, 22, 1, 0,
        0, 0, 24, 25, 1, 0, 0, 0, 25, 3, 1, 0, 0, 0, 26, 28, 3, 8, 4, 0, 27, 29, 3, 2, 1, 0, 28, 27,
        1, 0, 0, 0, 28, 29, 1, 0, 0, 0, 29, 141, 1, 0, 0, 0, 30, 141, 3, 6, 3, 0, 31, 33, 5, 12, 0,
        0, 32, 34, 3, 2, 1, 0, 33, 32, 1, 0, 0, 0, 33, 34, 1, 0, 0, 0, 34, 35, 1, 0, 0, 0, 35, 41,
        6, 2, -1, 0, 36, 40, 3, 4, 2, 0, 37, 40, 5, 69, 0, 0, 38, 40, 5, 1, 0, 0, 39, 36, 1, 0, 0,
        0, 39, 37, 1, 0, 0, 0, 39, 38, 1, 0, 0, 0, 40, 43, 1, 0, 0, 0, 41, 39, 1, 0, 0, 0, 41, 42,
        1, 0, 0, 0, 42, 44, 1, 0, 0, 0, 43, 41, 1, 0, 0, 0, 44, 45, 6, 2, -
        1, 0, 45, 141, 5, 13, 0,
        0, 46, 51, 5, 16, 0, 0, 47, 50, 3, 4, 2, 0, 48, 50, 3, 2, 1, 0, 49, 47, 1, 0, 0, 0, 49, 48,
        1, 0, 0, 0, 50, 53, 1, 0, 0, 0, 51, 49, 1, 0, 0, 0, 51, 52, 1, 0, 0, 0, 52, 54, 1, 0, 0, 0,
        53, 51, 1, 0, 0, 0, 54, 141, 5, 17, 0, 0, 55, 60, 5, 4, 0, 0, 56, 59, 3, 4, 2, 0, 57, 59,
        3, 2, 1, 0, 58, 56, 1, 0, 0, 0, 58, 57, 1, 0, 0, 0, 59, 62, 1, 0, 0, 0, 60, 58, 1, 0, 0, 0,
        60, 61, 1, 0, 0, 0, 61, 63, 1, 0, 0, 0, 62, 60, 1, 0, 0, 0, 63, 141, 5, 3, 0, 0, 64, 69, 5,
        2, 0, 0, 65, 68, 3, 4, 2, 0, 66, 68, 3, 2, 1, 0, 67, 65, 1, 0, 0, 0, 67, 66, 1, 0, 0, 0, 68,
        71, 1, 0, 0, 0, 69, 67, 1, 0, 0, 0, 69, 70, 1, 0, 0, 0, 70, 72, 1, 0, 0, 0, 71, 69, 1, 0, 0,
        0, 72, 73, 5, 3, 0, 0, 73, 141, 5, 3, 0, 0, 74, 76, 5, 44, 0, 0, 75, 77, 3, 2, 1, 0, 76, 75,
        1, 0, 0, 0, 76, 77, 1, 0, 0, 0, 77, 78, 1, 0, 0, 0, 78, 83, 5, 16, 0, 0, 79, 82, 3, 4, 2, 0,
        80, 82, 3, 2, 1, 0, 81, 79, 1, 0, 0, 0, 81, 80, 1, 0, 0, 0, 82, 85, 1, 0, 0, 0, 83, 81, 1,
        0, 0, 0, 83, 84, 1, 0, 0, 0, 84, 86, 1, 0, 0, 0, 85, 83, 1, 0, 0, 0, 86, 88, 5, 17, 0, 0, 87,
        89, 3, 2, 1, 0, 88, 87, 1, 0, 0, 0, 88, 89, 1, 0, 0, 0, 89, 141, 1, 0, 0, 0, 90, 92, 5, 33,
        0, 0, 91, 93, 3, 2, 1, 0, 92, 91, 1, 0, 0, 0, 92, 93, 1, 0, 0, 0, 93, 94, 1, 0, 0, 0, 94, 99,
        5, 16, 0, 0, 95, 98, 3, 4, 2, 0, 96, 98, 3, 2, 1, 0, 97, 95, 1, 0, 0, 0, 97, 96, 1, 0, 0, 0,
        98, 101, 1, 0, 0, 0, 99, 97, 1, 0, 0, 0, 99, 100, 1, 0, 0, 0, 100, 102, 1, 0, 0, 0, 101,
        99, 1, 0, 0, 0, 102, 104, 5, 17, 0, 0, 103, 105, 3, 2, 1, 0, 104, 103, 1, 0, 0, 0, 104,
        105, 1, 0, 0, 0, 105, 141, 1, 0, 0, 0, 106, 108, 5, 32, 0, 0, 107, 109, 3, 2, 1, 0, 108,
        107, 1, 0, 0, 0, 108, 109, 1, 0, 0, 0, 109, 110, 1, 0, 0, 0, 110, 115, 5, 16, 0, 0, 111,
        114, 3, 4, 2, 0, 112, 114, 3, 2, 1, 0, 113, 111, 1, 0, 0, 0, 113, 112, 1, 0, 0, 0, 114,
        117, 1, 0, 0, 0, 115, 113, 1, 0, 0, 0, 115, 116, 1, 0, 0, 0, 116, 118, 1, 0, 0, 0, 117,
        115, 1, 0, 0, 0, 118, 120, 5, 17, 0, 0, 119, 121, 3, 2, 1, 0, 120, 119, 1, 0, 0, 0, 120,
        121, 1, 0, 0, 0, 121, 141, 1, 0, 0, 0, 122, 124, 5, 35, 0, 0, 123, 125, 3, 2, 1, 0, 124,
        123, 1, 0, 0, 0, 124, 125, 1, 0, 0, 0, 125, 126, 1, 0, 0, 0, 126, 131, 5, 16, 0, 0, 127,
        130, 3, 4, 2, 0, 128, 130, 3, 2, 1, 0, 129, 127, 1, 0, 0, 0, 129, 128, 1, 0, 0, 0, 130,
        133, 1, 0, 0, 0, 131, 129, 1, 0, 0, 0, 131, 132, 1, 0, 0, 0, 132, 134, 1, 0, 0, 0, 133,
        131, 1, 0, 0, 0, 134, 136, 5, 17, 0, 0, 135, 137, 3, 2, 1, 0, 136, 135, 1, 0, 0, 0, 136,
        137, 1, 0, 0, 0, 137, 141, 1, 0, 0, 0, 138, 139, 5, 36, 0, 0, 139, 141, 6, 2, -1, 0, 140,
        26, 1, 0, 0, 0, 140, 30, 1, 0, 0, 0, 140, 31, 1, 0, 0, 0, 140, 46, 1, 0, 0, 0, 140, 55, 1,
        0, 0, 0, 140, 64, 1, 0, 0, 0, 140, 74, 1, 0, 0, 0, 140, 90, 1, 0, 0, 0, 140, 106, 1, 0, 0,
        0, 140, 122, 1, 0, 0, 0, 140, 138, 1, 0, 0, 0, 141, 5, 1, 0, 0, 0, 142, 143, 7, 0, 0, 0,
        143, 7, 1, 0, 0, 0, 144, 145, 7, 1, 0, 0, 145, 9, 1, 0, 0, 0, 30, 13, 15, 24, 28, 33, 39,
        41, 49, 51, 58, 60, 67, 69, 76, 81, 83, 88, 92, 97, 99, 104, 108, 113, 115, 120, 124,
        129, 131, 136, 140
    ]


class RFilter (Parser):

    grammarFileName = "RFilter.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = ["<INVALID>", "';'", "'[['", "']'", "'['", "'-'", "'+'",
                    "'!'", "'&'", "'&&'", "'|'", "'||'", "'{'", "'}'",
                    "'?'", "'break'", "'('", "')'", "'.'", "','", "'='",
                    "'...'", "'NULL'", "':'", "'%in%'", "'NA'", "'Inf'",
                    "'NaN'", "'TRUE'", "'FALSE'", "'next'", "'repeat'",
                    "'while'", "'for'", "'in'", "'if'", "'else'", "'return'",
                    "'within'", "'<-'", "'<<-'", "'->'", "'->>'", "':='",
                    "'function'", "'~'", "'::'", "':::'", "'$'", "'@'",
                    "'*'", "'/'", "'%'", "'^'", "'%%'", "'>'", "'>='",
                    "'<'", "'<='", "'=='", "'!='"]

    symbolicNames = ["<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                     "<INVALID>", "PARENTHESIS", "HEX", "INT", "FLOAT",
                     "COMPLEX", "STRING", "ID", "USER_OP", "NL", "WS"]

    RULE_stream = 0
    RULE_eat = 1
    RULE_elem = 2
    RULE_atom = 3
    RULE_op = 4

    ruleNames = ["stream", "eat", "elem", "atom", "op"]

    EOF = Token.EOF
    T__0 = 1
    T__1 = 2
    T__2 = 3
    T__3 = 4
    T__4 = 5
    T__5 = 6
    T__6 = 7
    T__7 = 8
    T__8 = 9
    T__9 = 10
    T__10 = 11
    T__11 = 12
    T__12 = 13
    T__13 = 14
    T__14 = 15
    T__15 = 16
    T__16 = 17
    T__17 = 18
    T__18 = 19
    T__19 = 20
    T__20 = 21
    T__21 = 22
    T__22 = 23
    T__23 = 24
    T__24 = 25
    T__25 = 26
    T__26 = 27
    T__27 = 28
    T__28 = 29
    T__29 = 30
    T__30 = 31
    T__31 = 32
    T__32 = 33
    T__33 = 34
    T__34 = 35
    T__35 = 36
    T__36 = 37
    T__37 = 38
    T__38 = 39
    T__39 = 40
    T__40 = 41
    T__41 = 42
    T__42 = 43
    T__43 = 44
    T__44 = 45
    T__45 = 46
    T__46 = 47
    T__47 = 48
    T__48 = 49
    T__49 = 50
    T__50 = 51
    T__51 = 52
    T__52 = 53
    T__53 = 54
    T__54 = 55
    T__55 = 56
    T__56 = 57
    T__57 = 58
    T__58 = 59
    T__59 = 60
    PARENTHESIS = 61
    HEX = 62
    INT = 63
    FLOAT = 64
    COMPLEX = 65
    STRING = 66
    ID = 67
    USER_OP = 68
    NL = 69
    WS = 70

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.10.1")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None

    curlies = 0

    class StreamContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(RFilter.EOF, 0)

        def elem(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(RFilter.ElemContext)
            else:
                return self.getTypedRuleContext(RFilter.ElemContext, i)

        def NL(self, i: int = None):
            if i is None:
                return self.getTokens(RFilter.NL)
            else:
                return self.getToken(RFilter.NL, i)

        def getRuleIndex(self):
            return RFilter.RULE_stream

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStream"):
                listener.enterStream(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStream"):
                listener.exitStream(self)

    def stream(self):

        localctx = RFilter.StreamContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_stream)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 15
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__0) | (1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                self.state = 13
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                    self.state = 10
                    self.elem()
                    pass
                elif token in [RFilter.NL]:
                    self.state = 11
                    self.match(RFilter.NL)
                    pass
                elif token in [RFilter.T__0]:
                    self.state = 12
                    self.match(RFilter.T__0)
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 17
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 18
            self.match(RFilter.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class EatContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self._NL = None  # Token

        def NL(self, i: int = None):
            if i is None:
                return self.getTokens(RFilter.NL)
            else:
                return self.getToken(RFilter.NL, i)

        def getRuleIndex(self):
            return RFilter.RULE_eat

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEat"):
                listener.enterEat(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEat"):
                listener.exitEat(self)

    def eat(self):

        localctx = RFilter.EatContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_eat)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 22
            self._errHandler.sync(self)
            _alt = 1
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 20
                    localctx._NL = self.match(RFilter.NL)
                    localctx._NL.setChannel(Token.HIDDEN_CHANNEL)

                else:
                    raise NoViableAltException(self)
                self.state = 24
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 2, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ElemContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def op(self):
            return self.getTypedRuleContext(RFilter.OpContext, 0)

        def eat(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(RFilter.EatContext)
            else:
                return self.getTypedRuleContext(RFilter.EatContext, i)

        def atom(self):
            return self.getTypedRuleContext(RFilter.AtomContext, 0)

        def elem(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(RFilter.ElemContext)
            else:
                return self.getTypedRuleContext(RFilter.ElemContext, i)

        def NL(self, i: int = None):
            if i is None:
                return self.getTokens(RFilter.NL)
            else:
                return self.getToken(RFilter.NL, i)

        def getRuleIndex(self):
            return RFilter.RULE_elem

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterElem"):
                listener.enterElem(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitElem"):
                listener.exitElem(self)

    def elem(self):

        localctx = RFilter.ElemContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_elem)
        self._la = 0  # Token type
        try:
            self.state = 140
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__13, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__22, RFilter.T__30, RFilter.T__33, RFilter.T__38, RFilter.T__40, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.USER_OP]:
                self.enterOuterAlt(localctx, 1)
                self.state = 26
                self.op()
                self.state = 28
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 3, self._ctx)
                if la_ == 1:
                    self.state = 27
                    self.eat()

                pass
            elif token in [RFilter.T__14, RFilter.T__21, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID]:
                self.enterOuterAlt(localctx, 2)
                self.state = 30
                self.atom()
                pass
            elif token in [RFilter.T__11]:
                self.enterOuterAlt(localctx, 3)
                self.state = 31
                self.match(RFilter.T__11)
                self.state = 33
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 4, self._ctx)
                if la_ == 1:
                    self.state = 32
                    self.eat()

                curlies += 1
                self.state = 41
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__0) | (1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                    self.state = 39
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                        self.state = 36
                        self.elem()
                        pass
                    elif token in [RFilter.NL]:
                        self.state = 37
                        self.match(RFilter.NL)
                        pass
                    elif token in [RFilter.T__0]:
                        self.state = 38
                        self.match(RFilter.T__0)
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 43
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                curlies -= 1
                self.state = 45
                self.match(RFilter.T__12)
                pass
            elif token in [RFilter.T__15]:
                self.enterOuterAlt(localctx, 4)
                self.state = 46
                self.match(RFilter.T__15)
                self.state = 51
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                    self.state = 49
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                        self.state = 47
                        self.elem()
                        pass
                    elif token in [RFilter.NL]:
                        self.state = 48
                        self.eat()
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 53
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 54
                self.match(RFilter.T__16)
                pass
            elif token in [RFilter.T__3]:
                self.enterOuterAlt(localctx, 5)
                self.state = 55
                self.match(RFilter.T__3)
                self.state = 60
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                    self.state = 58
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                        self.state = 56
                        self.elem()
                        pass
                    elif token in [RFilter.NL]:
                        self.state = 57
                        self.eat()
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 62
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 63
                self.match(RFilter.T__2)
                pass
            elif token in [RFilter.T__1]:
                self.enterOuterAlt(localctx, 6)
                self.state = 64
                self.match(RFilter.T__1)
                self.state = 69
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                    self.state = 67
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                        self.state = 65
                        self.elem()
                        pass
                    elif token in [RFilter.NL]:
                        self.state = 66
                        self.eat()
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 71
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 72
                self.match(RFilter.T__2)
                self.state = 73
                self.match(RFilter.T__2)
                pass
            elif token in [RFilter.T__43]:
                self.enterOuterAlt(localctx, 7)
                self.state = 74
                self.match(RFilter.T__43)
                self.state = 76
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == RFilter.NL:
                    self.state = 75
                    self.eat()

                self.state = 78
                self.match(RFilter.T__15)
                self.state = 83
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                    self.state = 81
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                        self.state = 79
                        self.elem()
                        pass
                    elif token in [RFilter.NL]:
                        self.state = 80
                        self.eat()
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 85
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 86
                self.match(RFilter.T__16)
                self.state = 88
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 16, self._ctx)
                if la_ == 1:
                    self.state = 87
                    self.eat()

                pass
            elif token in [RFilter.T__32]:
                self.enterOuterAlt(localctx, 8)
                self.state = 90
                self.match(RFilter.T__32)
                self.state = 92
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == RFilter.NL:
                    self.state = 91
                    self.eat()

                self.state = 94
                self.match(RFilter.T__15)
                self.state = 99
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                    self.state = 97
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                        self.state = 95
                        self.elem()
                        pass
                    elif token in [RFilter.NL]:
                        self.state = 96
                        self.eat()
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 101
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 102
                self.match(RFilter.T__16)
                self.state = 104
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 20, self._ctx)
                if la_ == 1:
                    self.state = 103
                    self.eat()

                pass
            elif token in [RFilter.T__31]:
                self.enterOuterAlt(localctx, 9)
                self.state = 106
                self.match(RFilter.T__31)
                self.state = 108
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == RFilter.NL:
                    self.state = 107
                    self.eat()

                self.state = 110
                self.match(RFilter.T__15)
                self.state = 115
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                    self.state = 113
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                        self.state = 111
                        self.elem()
                        pass
                    elif token in [RFilter.NL]:
                        self.state = 112
                        self.eat()
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 117
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 118
                self.match(RFilter.T__16)
                self.state = 120
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 24, self._ctx)
                if la_ == 1:
                    self.state = 119
                    self.eat()

                pass
            elif token in [RFilter.T__34]:
                self.enterOuterAlt(localctx, 10)
                self.state = 122
                self.match(RFilter.T__34)
                self.state = 124
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == RFilter.NL:
                    self.state = 123
                    self.eat()

                self.state = 126
                self.match(RFilter.T__15)
                self.state = 131
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << RFilter.T__1) | (1 << RFilter.T__3) | (1 << RFilter.T__4) | (1 << RFilter.T__5) | (1 << RFilter.T__6) | (1 << RFilter.T__7) | (1 << RFilter.T__8) | (1 << RFilter.T__9) | (1 << RFilter.T__10) | (1 << RFilter.T__11) | (1 << RFilter.T__13) | (1 << RFilter.T__14) | (1 << RFilter.T__15) | (1 << RFilter.T__18) | (1 << RFilter.T__19) | (1 << RFilter.T__20) | (1 << RFilter.T__21) | (1 << RFilter.T__22) | (1 << RFilter.T__24) | (1 << RFilter.T__25) | (1 << RFilter.T__26) | (1 << RFilter.T__27) | (1 << RFilter.T__28) | (1 << RFilter.T__29) | (1 << RFilter.T__30) | (1 << RFilter.T__31) | (1 << RFilter.T__32) | (1 << RFilter.T__33) | (1 << RFilter.T__34) | (1 << RFilter.T__35) | (1 << RFilter.T__38) | (1 << RFilter.T__40) | (1 << RFilter.T__43) | (1 << RFilter.T__44) | (1 << RFilter.T__45) | (1 << RFilter.T__46) | (1 << RFilter.T__47) | (1 << RFilter.T__48) | (1 << RFilter.T__49) | (1 << RFilter.T__50) | (1 << RFilter.T__52) | (1 << RFilter.T__54) | (1 << RFilter.T__55) | (1 << RFilter.T__56) | (1 << RFilter.T__57) | (1 << RFilter.T__58) | (1 << RFilter.T__59) | (1 << RFilter.HEX) | (1 << RFilter.INT))) != 0) or ((((_la - 64)) & ~0x3f) == 0 and ((1 << (_la - 64)) & ((1 << (RFilter.FLOAT - 64)) | (1 << (RFilter.COMPLEX - 64)) | (1 << (RFilter.STRING - 64)) | (1 << (RFilter.ID - 64)) | (1 << (RFilter.USER_OP - 64)) | (1 << (RFilter.NL - 64)))) != 0):
                    self.state = 129
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [RFilter.T__1, RFilter.T__3, RFilter.T__4, RFilter.T__5, RFilter.T__6, RFilter.T__7, RFilter.T__8, RFilter.T__9, RFilter.T__10, RFilter.T__11, RFilter.T__13, RFilter.T__14, RFilter.T__15, RFilter.T__18, RFilter.T__19, RFilter.T__20, RFilter.T__21, RFilter.T__22, RFilter.T__24, RFilter.T__25, RFilter.T__26, RFilter.T__27, RFilter.T__28, RFilter.T__29, RFilter.T__30, RFilter.T__31, RFilter.T__32, RFilter.T__33, RFilter.T__34, RFilter.T__35, RFilter.T__38, RFilter.T__40, RFilter.T__43, RFilter.T__44, RFilter.T__45, RFilter.T__46, RFilter.T__47, RFilter.T__48, RFilter.T__49, RFilter.T__50, RFilter.T__52, RFilter.T__54, RFilter.T__55, RFilter.T__56, RFilter.T__57, RFilter.T__58, RFilter.T__59, RFilter.HEX, RFilter.INT, RFilter.FLOAT, RFilter.COMPLEX, RFilter.STRING, RFilter.ID, RFilter.USER_OP]:
                        self.state = 127
                        self.elem()
                        pass
                    elif token in [RFilter.NL]:
                        self.state = 128
                        self.eat()
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 133
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 134
                self.match(RFilter.T__16)
                self.state = 136
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 28, self._ctx)
                if la_ == 1:
                    self.state = 135
                    self.eat()

                pass
            elif token in [RFilter.T__35]:
                self.enterOuterAlt(localctx, 11)
                self.state = 138
                self.match(RFilter.T__35)

                # ``inside a compound expression, a newline before else is discarded,
                # whereas at the outermost level, the newline terminates the if
                # construction and a subsequent else causes a syntax error.''
                # /*
                # Works here
                #     if (1 == 0) {print(1)} else {print(2)}
                #
                # and correctly gets error here:
                #
                #     if (1 == 0) {print(1)}
                #     else {print(2)}
                #
                # this works too:
                #
                #     if (1 == 0) {
                #         if (2 == 0) print(1)
                #         else print(2)
                #     }
                # */
                tok = _input.LT(-2)
                if curlies > 0 and tok.getType() == NL:
                    tok.setChannel(Token.HIDDEN_CHANNEL)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AtomContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(RFilter.ID, 0)

        def STRING(self):
            return self.getToken(RFilter.STRING, 0)

        def HEX(self):
            return self.getToken(RFilter.HEX, 0)

        def INT(self):
            return self.getToken(RFilter.INT, 0)

        def FLOAT(self):
            return self.getToken(RFilter.FLOAT, 0)

        def COMPLEX(self):
            return self.getToken(RFilter.COMPLEX, 0)

        def getRuleIndex(self):
            return RFilter.RULE_atom

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAtom"):
                listener.enterAtom(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAtom"):
                listener.exitAtom(self)

    def atom(self):

        localctx = RFilter.AtomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_atom)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 142
            _la = self._input.LA(1)
            if not (((((_la - 15)) & ~0x3f) == 0 and ((1 << (_la - 15)) & ((1 << (RFilter.T__14 - 15)) | (1 << (RFilter.T__21 - 15)) | (1 << (RFilter.T__24 - 15)) | (1 << (RFilter.T__25 - 15)) | (1 << (RFilter.T__26 - 15)) | (1 << (RFilter.T__27 - 15)) | (1 << (RFilter.T__28 - 15)) | (1 << (RFilter.T__29 - 15)) | (1 << (RFilter.HEX - 15)) | (1 << (RFilter.INT - 15)) | (1 << (RFilter.FLOAT - 15)) | (1 << (RFilter.COMPLEX - 15)) | (1 << (RFilter.STRING - 15)) | (1 << (RFilter.ID - 15)))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OpContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def USER_OP(self):
            return self.getToken(RFilter.USER_OP, 0)

        def getRuleIndex(self):
            return RFilter.RULE_op

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOp"):
                listener.enterOp(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOp"):
                listener.exitOp(self)

    def op(self):

        localctx = RFilter.OpContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_op)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 144
            _la = self._input.LA(1)
            if not (((((_la - 5)) & ~0x3f) == 0 and ((1 << (_la - 5)) & ((1 << (RFilter.T__4 - 5)) | (1 << (RFilter.T__5 - 5)) | (1 << (RFilter.T__6 - 5)) | (1 << (RFilter.T__7 - 5)) | (1 << (RFilter.T__8 - 5)) | (1 << (RFilter.T__9 - 5)) | (1 << (RFilter.T__10 - 5)) | (1 << (RFilter.T__13 - 5)) | (1 << (RFilter.T__18 - 5)) | (1 << (RFilter.T__19 - 5)) | (1 << (RFilter.T__20 - 5)) | (1 << (RFilter.T__22 - 5)) | (1 << (RFilter.T__30 - 5)) | (1 << (RFilter.T__33 - 5)) | (1 << (RFilter.T__38 - 5)) | (1 << (RFilter.T__40 - 5)) | (1 << (RFilter.T__44 - 5)) | (1 << (RFilter.T__45 - 5)) | (1 << (RFilter.T__46 - 5)) | (1 << (RFilter.T__47 - 5)) | (1 << (RFilter.T__48 - 5)) | (1 << (RFilter.T__49 - 5)) | (1 << (RFilter.T__50 - 5)) | (1 << (RFilter.T__52 - 5)) | (1 << (RFilter.T__54 - 5)) | (1 << (RFilter.T__55 - 5)) | (1 << (RFilter.T__56 - 5)) | (1 << (RFilter.T__57 - 5)) | (1 << (RFilter.T__58 - 5)) | (1 << (RFilter.T__59 - 5)) | (1 << (RFilter.USER_OP - 5)))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



"""
@file

@brief defines logged exceptions for SQL requests
"""


class DBException(Exception):

    """
    custom exception
    """
    pass


class ExceptionSQL(DBException):

    """
    exception related to SQL instructions and only them
    """

    def __init__(self, description, ex, sql, log=True):
        """
        @param      description     message
        @param      ex              sqlite exception
        @param      sql             SQL instruction
        @param      log             log the exception
        """
        DBException.__init__(self, description + "\n" + sql)
        self.ex = ex
        self.sql = sql
        if log:
            print(description + "\n" + sql)  # pragma: no cover

    def __str__(self):
        mes = Exception.__str__(self)
        mes += "\n" + str(self.ex)
        mes += "\n" + "\n".join(repr(self.sql).split("\\n"))
        if len(mes) > 10000:
            mes = mes[:10000] + "\n..."  # pragma: no cover
        return mes



"""
@file
@brief Plotting maps.
"""


def plot_map_france(ax=None, scale='50m'):
    """
    Creates a map for France using :epkg:`cartopy`.

    @param      ax          existing axes or None to create ones
    @param      scale       scale in (`10m`, `50m`, `110m`)
    @return                 ax

    .. plot::

        from matplotlib import pyplot as plt
        from pyensae.graphhelper import plot_map_france

        plot_map_france()
        plt.show()

    You may want to add `ax.set_extent([-5., 10., 38., 52.])`
    after `ax = plot_map_france()`.
    See also example :ref:`l-map-france`.
    """

    if ax is None:  # pragma: no cover
        import matplotlib.pyplot as plt  # pylint: disable=C0415
        import cartopy.crs as ccrs  # pylint: disable=C0415
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
        # The following line may make the program crash on debian + python3.9.1.
        # ax.set_extent([-5, 10, 38, 52])

    import cartopy.feature as cfeature  # pylint: disable=C0415
    ax.add_feature(cfeature.OCEAN.with_scale(scale))
    ax.add_feature(cfeature.RIVERS.with_scale(scale))
    ax.add_feature(cfeature.BORDERS.with_scale(scale), linestyle=':')
    return ax


def plot_map_france_polygon(geometry, colors, ax=None, scale='50m'):
    """
    Plots polygons into a map for France.

    @param      geometry        series of polygons
    @param      colors          colors
    @param      scale           scale, see @see fn map_france
    @param      ax              existing axes, None to create one
    @return                     ax

    .. plot::

        from matplotlib import pyplot as plt
        import cartopy.crs as ccrs
        from pyensae.datasource import load_french_departements
        from pyensae.graphhelper import plot_map_france, plot_map_france_polygon

        # loads the French departments
        df = load_french_departements()

        fig = plt.figure(figsize=(7,7))
        ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
        N = float(df.shape[0])
        plot_map_france_polygon(
            ax=ax, geometry=df['geometry'],
            colors=[(i/N, i/N, i/N) for i in range(df.shape[0])])

        plt.show()

    You may want to add `ax.set_extent([-5., 10., 38., 52.])`
    after `ax = plot_map_france_polygon()`.
    See also example :ref:`l-map-france`.
    """
    from geopandas.plotting import plot_polygon_collection  # pylint: disable=C0415
    ax = plot_map_france(scale=scale, ax=ax)
    plot_polygon_collection(
        ax, geometry, facecolor=colors, values=None, edgecolor='black')
    return ax



# Generated from \SQLite.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .SQLiteParser import SQLiteParser
else:
    from SQLiteParser import SQLiteParser

# This class defines a complete listener for a parse tree produced by SQLiteParser.


class SQLiteListener(ParseTreeListener):

    # Enter a parse tree produced by SQLiteParser#parse.
    def enterParse(self, ctx: SQLiteParser.ParseContext):
        pass

    # Exit a parse tree produced by SQLiteParser#parse.
    def exitParse(self, ctx: SQLiteParser.ParseContext):
        pass

    # Enter a parse tree produced by SQLiteParser#error.
    def enterError(self, ctx):
        pass

    # Exit a parse tree produced by SQLiteParser#error.
    def exitError(self, ctx):
        pass

    # Enter a parse tree produced by SQLiteParser#sql_stmt_list.
    def enterSql_stmt_list(self, ctx: SQLiteParser.Sql_stmt_listContext):
        pass

    # Exit a parse tree produced by SQLiteParser#sql_stmt_list.
    def exitSql_stmt_list(self, ctx: SQLiteParser.Sql_stmt_listContext):
        pass

    # Enter a parse tree produced by SQLiteParser#sql_stmt.
    def enterSql_stmt(self, ctx: SQLiteParser.Sql_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#sql_stmt.
    def exitSql_stmt(self, ctx: SQLiteParser.Sql_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#alter_table_stmt.
    def enterAlter_table_stmt(self, ctx: SQLiteParser.Alter_table_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#alter_table_stmt.
    def exitAlter_table_stmt(self, ctx: SQLiteParser.Alter_table_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#analyze_stmt.
    def enterAnalyze_stmt(self, ctx: SQLiteParser.Analyze_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#analyze_stmt.
    def exitAnalyze_stmt(self, ctx: SQLiteParser.Analyze_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#attach_stmt.
    def enterAttach_stmt(self, ctx: SQLiteParser.Attach_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#attach_stmt.
    def exitAttach_stmt(self, ctx: SQLiteParser.Attach_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#begin_stmt.
    def enterBegin_stmt(self, ctx: SQLiteParser.Begin_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#begin_stmt.
    def exitBegin_stmt(self, ctx: SQLiteParser.Begin_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#commit_stmt.
    def enterCommit_stmt(self, ctx: SQLiteParser.Commit_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#commit_stmt.
    def exitCommit_stmt(self, ctx: SQLiteParser.Commit_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#compound_select_stmt.
    def enterCompound_select_stmt(self, ctx: SQLiteParser.Compound_select_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#compound_select_stmt.
    def exitCompound_select_stmt(self, ctx: SQLiteParser.Compound_select_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#create_index_stmt.
    def enterCreate_index_stmt(self, ctx: SQLiteParser.Create_index_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#create_index_stmt.
    def exitCreate_index_stmt(self, ctx: SQLiteParser.Create_index_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#create_table_stmt.
    def enterCreate_table_stmt(self, ctx: SQLiteParser.Create_table_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#create_table_stmt.
    def exitCreate_table_stmt(self, ctx: SQLiteParser.Create_table_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#create_trigger_stmt.
    def enterCreate_trigger_stmt(self, ctx: SQLiteParser.Create_trigger_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#create_trigger_stmt.
    def exitCreate_trigger_stmt(self, ctx: SQLiteParser.Create_trigger_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#create_view_stmt.
    def enterCreate_view_stmt(self, ctx: SQLiteParser.Create_view_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#create_view_stmt.
    def exitCreate_view_stmt(self, ctx: SQLiteParser.Create_view_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#create_virtual_table_stmt.
    def enterCreate_virtual_table_stmt(self, ctx: SQLiteParser.Create_virtual_table_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#create_virtual_table_stmt.
    def exitCreate_virtual_table_stmt(self, ctx: SQLiteParser.Create_virtual_table_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#delete_stmt.
    def enterDelete_stmt(self, ctx: SQLiteParser.Delete_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#delete_stmt.
    def exitDelete_stmt(self, ctx: SQLiteParser.Delete_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#delete_stmt_limited.
    def enterDelete_stmt_limited(self, ctx: SQLiteParser.Delete_stmt_limitedContext):
        pass

    # Exit a parse tree produced by SQLiteParser#delete_stmt_limited.
    def exitDelete_stmt_limited(self, ctx: SQLiteParser.Delete_stmt_limitedContext):
        pass

    # Enter a parse tree produced by SQLiteParser#detach_stmt.
    def enterDetach_stmt(self, ctx: SQLiteParser.Detach_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#detach_stmt.
    def exitDetach_stmt(self, ctx: SQLiteParser.Detach_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#drop_index_stmt.
    def enterDrop_index_stmt(self, ctx):
        pass

    # Exit a parse tree produced by SQLiteParser#drop_index_stmt.
    def exitDrop_index_stmt(self, ctx):
        pass

    # Enter a parse tree produced by SQLiteParser#drop_table_stmt.
    def enterDrop_table_stmt(self, ctx):
        pass

    # Exit a parse tree produced by SQLiteParser#drop_table_stmt.
    def exitDrop_table_stmt(self, ctx):
        pass

    # Enter a parse tree produced by SQLiteParser#drop_trigger_stmt.
    def enterDrop_trigger_stmt(self, ctx):
        pass

    # Exit a parse tree produced by SQLiteParser#drop_trigger_stmt.
    def exitDrop_trigger_stmt(self, ctx):
        pass

    # Enter a parse tree produced by SQLiteParser#drop_view_stmt.
    def enterDrop_view_stmt(self, ctx):
        pass

    # Exit a parse tree produced by SQLiteParser#drop_view_stmt.
    def exitDrop_view_stmt(self, ctx):
        pass

    # Enter a parse tree produced by SQLiteParser#factored_select_stmt.
    def enterFactored_select_stmt(self, ctx: SQLiteParser.Factored_select_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#factored_select_stmt.
    def exitFactored_select_stmt(self, ctx: SQLiteParser.Factored_select_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#insert_stmt.
    def enterInsert_stmt(self, ctx: SQLiteParser.Insert_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#insert_stmt.
    def exitInsert_stmt(self, ctx: SQLiteParser.Insert_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#pragma_stmt.
    def enterPragma_stmt(self, ctx: SQLiteParser.Pragma_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#pragma_stmt.
    def exitPragma_stmt(self, ctx: SQLiteParser.Pragma_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#reindex_stmt.
    def enterReindex_stmt(self, ctx: SQLiteParser.Reindex_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#reindex_stmt.
    def exitReindex_stmt(self, ctx: SQLiteParser.Reindex_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#release_stmt.
    def enterRelease_stmt(self, ctx: SQLiteParser.Release_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#release_stmt.
    def exitRelease_stmt(self, ctx: SQLiteParser.Release_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#rollback_stmt.
    def enterRollback_stmt(self, ctx: SQLiteParser.Rollback_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#rollback_stmt.
    def exitRollback_stmt(self, ctx: SQLiteParser.Rollback_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#savepoint_stmt.
    def enterSavepoint_stmt(self, ctx: SQLiteParser.Savepoint_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#savepoint_stmt.
    def exitSavepoint_stmt(self, ctx: SQLiteParser.Savepoint_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#simple_select_stmt.
    def enterSimple_select_stmt(self, ctx: SQLiteParser.Simple_select_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#simple_select_stmt.
    def exitSimple_select_stmt(self, ctx: SQLiteParser.Simple_select_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#select_stmt.
    def enterSelect_stmt(self, ctx: SQLiteParser.Select_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#select_stmt.
    def exitSelect_stmt(self, ctx: SQLiteParser.Select_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#select_or_values.
    def enterSelect_or_values(self, ctx):
        pass

    # Exit a parse tree produced by SQLiteParser#select_or_values.
    def exitSelect_or_values(self, ctx):
        pass

    # Enter a parse tree produced by SQLiteParser#update_stmt.
    def enterUpdate_stmt(self, ctx: SQLiteParser.Update_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#update_stmt.
    def exitUpdate_stmt(self, ctx: SQLiteParser.Update_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#update_stmt_limited.
    def enterUpdate_stmt_limited(self, ctx: SQLiteParser.Update_stmt_limitedContext):
        pass

    # Exit a parse tree produced by SQLiteParser#update_stmt_limited.
    def exitUpdate_stmt_limited(self, ctx: SQLiteParser.Update_stmt_limitedContext):
        pass

    # Enter a parse tree produced by SQLiteParser#vacuum_stmt.
    def enterVacuum_stmt(self, ctx: SQLiteParser.Vacuum_stmtContext):
        pass

    # Exit a parse tree produced by SQLiteParser#vacuum_stmt.
    def exitVacuum_stmt(self, ctx: SQLiteParser.Vacuum_stmtContext):
        pass

    # Enter a parse tree produced by SQLiteParser#column_def.
    def enterColumn_def(self, ctx: SQLiteParser.Column_defContext):
        pass

    # Exit a parse tree produced by SQLiteParser#column_def.
    def exitColumn_def(self, ctx: SQLiteParser.Column_defContext):
        pass

    # Enter a parse tree produced by SQLiteParser#type_name.
    def enterType_name(self, ctx: SQLiteParser.Type_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#type_name.
    def exitType_name(self, ctx: SQLiteParser.Type_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#column_constraint.
    def enterColumn_constraint(self, ctx: SQLiteParser.Column_constraintContext):
        pass

    # Exit a parse tree produced by SQLiteParser#column_constraint.
    def exitColumn_constraint(self, ctx: SQLiteParser.Column_constraintContext):
        pass

    # Enter a parse tree produced by SQLiteParser#conflict_clause.
    def enterConflict_clause(self, ctx: SQLiteParser.Conflict_clauseContext):
        pass

    # Exit a parse tree produced by SQLiteParser#conflict_clause.
    def exitConflict_clause(self, ctx: SQLiteParser.Conflict_clauseContext):
        pass

    # Enter a parse tree produced by SQLiteParser#expr.
    def enterExpr(self, ctx: SQLiteParser.ExprContext):
        pass

    # Exit a parse tree produced by SQLiteParser#expr.
    def exitExpr(self, ctx: SQLiteParser.ExprContext):
        pass

    # Enter a parse tree produced by SQLiteParser#foreign_key_clause.
    def enterForeign_key_clause(self, ctx: SQLiteParser.Foreign_key_clauseContext):
        pass

    # Exit a parse tree produced by SQLiteParser#foreign_key_clause.
    def exitForeign_key_clause(self, ctx: SQLiteParser.Foreign_key_clauseContext):
        pass

    # Enter a parse tree produced by SQLiteParser#raise_function.
    def enterRaise_function(self, ctx: SQLiteParser.Raise_functionContext):
        pass

    # Exit a parse tree produced by SQLiteParser#raise_function.
    def exitRaise_function(self, ctx: SQLiteParser.Raise_functionContext):
        pass

    # Enter a parse tree produced by SQLiteParser#indexed_column.
    def enterIndexed_column(self, ctx: SQLiteParser.Indexed_columnContext):
        pass

    # Exit a parse tree produced by SQLiteParser#indexed_column.
    def exitIndexed_column(self, ctx: SQLiteParser.Indexed_columnContext):
        pass

    # Enter a parse tree produced by SQLiteParser#table_constraint.
    def enterTable_constraint(self, ctx: SQLiteParser.Table_constraintContext):
        pass

    # Exit a parse tree produced by SQLiteParser#table_constraint.
    def exitTable_constraint(self, ctx: SQLiteParser.Table_constraintContext):
        pass

    # Enter a parse tree produced by SQLiteParser#with_clause.
    def enterWith_clause(self, ctx: SQLiteParser.With_clauseContext):
        pass

    # Exit a parse tree produced by SQLiteParser#with_clause.
    def exitWith_clause(self, ctx: SQLiteParser.With_clauseContext):
        pass

    # Enter a parse tree produced by SQLiteParser#qualified_table_name.
    def enterQualified_table_name(self, ctx: SQLiteParser.Qualified_table_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#qualified_table_name.
    def exitQualified_table_name(self, ctx: SQLiteParser.Qualified_table_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#ordering_term.
    def enterOrdering_term(self, ctx: SQLiteParser.Ordering_termContext):
        pass

    # Exit a parse tree produced by SQLiteParser#ordering_term.
    def exitOrdering_term(self, ctx: SQLiteParser.Ordering_termContext):
        pass

    # Enter a parse tree produced by SQLiteParser#pragma_value.
    def enterPragma_value(self, ctx: SQLiteParser.Pragma_valueContext):
        pass

    # Exit a parse tree produced by SQLiteParser#pragma_value.
    def exitPragma_value(self, ctx: SQLiteParser.Pragma_valueContext):
        pass

    # Enter a parse tree produced by SQLiteParser#common_table_expression.
    def enterCommon_table_expression(self, ctx: SQLiteParser.Common_table_expressionContext):
        pass

    # Exit a parse tree produced by SQLiteParser#common_table_expression.
    def exitCommon_table_expression(self, ctx: SQLiteParser.Common_table_expressionContext):
        pass

    # Enter a parse tree produced by SQLiteParser#result_column.
    def enterResult_column(self, ctx: SQLiteParser.Result_columnContext):
        pass

    # Exit a parse tree produced by SQLiteParser#result_column.
    def exitResult_column(self, ctx: SQLiteParser.Result_columnContext):
        pass

    # Enter a parse tree produced by SQLiteParser#table_or_subquery.
    def enterTable_or_subquery(self, ctx: SQLiteParser.Table_or_subqueryContext):
        pass

    # Exit a parse tree produced by SQLiteParser#table_or_subquery.
    def exitTable_or_subquery(self, ctx: SQLiteParser.Table_or_subqueryContext):
        pass

    # Enter a parse tree produced by SQLiteParser#join_clause.
    def enterJoin_clause(self, ctx: SQLiteParser.Join_clauseContext):
        pass

    # Exit a parse tree produced by SQLiteParser#join_clause.
    def exitJoin_clause(self, ctx: SQLiteParser.Join_clauseContext):
        pass

    # Enter a parse tree produced by SQLiteParser#join_operator.
    def enterJoin_operator(self, ctx: SQLiteParser.Join_operatorContext):
        pass

    # Exit a parse tree produced by SQLiteParser#join_operator.
    def exitJoin_operator(self, ctx: SQLiteParser.Join_operatorContext):
        pass

    # Enter a parse tree produced by SQLiteParser#join_constraint.
    def enterJoin_constraint(self, ctx: SQLiteParser.Join_constraintContext):
        pass

    # Exit a parse tree produced by SQLiteParser#join_constraint.
    def exitJoin_constraint(self, ctx: SQLiteParser.Join_constraintContext):
        pass

    # Enter a parse tree produced by SQLiteParser#select_core.
    def enterSelect_core(self, ctx: SQLiteParser.Select_coreContext):
        pass

    # Exit a parse tree produced by SQLiteParser#select_core.
    def exitSelect_core(self, ctx: SQLiteParser.Select_coreContext):
        pass

    # Enter a parse tree produced by SQLiteParser#compound_operator.
    def enterCompound_operator(self, ctx: SQLiteParser.Compound_operatorContext):
        pass

    # Exit a parse tree produced by SQLiteParser#compound_operator.
    def exitCompound_operator(self, ctx: SQLiteParser.Compound_operatorContext):
        pass

    # Enter a parse tree produced by SQLiteParser#cte_table_name.
    def enterCte_table_name(self, ctx: SQLiteParser.Cte_table_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#cte_table_name.
    def exitCte_table_name(self, ctx: SQLiteParser.Cte_table_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#signed_number.
    def enterSigned_number(self, ctx: SQLiteParser.Signed_numberContext):
        pass

    # Exit a parse tree produced by SQLiteParser#signed_number.
    def exitSigned_number(self, ctx: SQLiteParser.Signed_numberContext):
        pass

    # Enter a parse tree produced by SQLiteParser#literal_value.
    def enterLiteral_value(self, ctx: SQLiteParser.Literal_valueContext):
        pass

    # Exit a parse tree produced by SQLiteParser#literal_value.
    def exitLiteral_value(self, ctx: SQLiteParser.Literal_valueContext):
        pass

    # Enter a parse tree produced by SQLiteParser#unary_operator.
    def enterUnary_operator(self, ctx: SQLiteParser.Unary_operatorContext):
        pass

    # Exit a parse tree produced by SQLiteParser#unary_operator.
    def exitUnary_operator(self, ctx: SQLiteParser.Unary_operatorContext):
        pass

    # Enter a parse tree produced by SQLiteParser#error_message.
    def enterError_message(self, ctx: SQLiteParser.Error_messageContext):
        pass

    # Exit a parse tree produced by SQLiteParser#error_message.
    def exitError_message(self, ctx: SQLiteParser.Error_messageContext):
        pass

    # Enter a parse tree produced by SQLiteParser#module_argument.
    def enterModule_argument(self, ctx: SQLiteParser.Module_argumentContext):
        pass

    # Exit a parse tree produced by SQLiteParser#module_argument.
    def exitModule_argument(self, ctx: SQLiteParser.Module_argumentContext):
        pass

    # Enter a parse tree produced by SQLiteParser#column_alias.
    def enterColumn_alias(self, ctx: SQLiteParser.Column_aliasContext):
        pass

    # Exit a parse tree produced by SQLiteParser#column_alias.
    def exitColumn_alias(self, ctx: SQLiteParser.Column_aliasContext):
        pass

    # Enter a parse tree produced by SQLiteParser#keyword.
    def enterKeyword(self, ctx: SQLiteParser.KeywordContext):
        pass

    # Exit a parse tree produced by SQLiteParser#keyword.
    def exitKeyword(self, ctx: SQLiteParser.KeywordContext):
        pass

    # Enter a parse tree produced by SQLiteParser#name.
    def enterName(self, ctx: SQLiteParser.NameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#name.
    def exitName(self, ctx: SQLiteParser.NameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#function_name.
    def enterFunction_name(self, ctx: SQLiteParser.Function_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#function_name.
    def exitFunction_name(self, ctx: SQLiteParser.Function_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#database_name.
    def enterDatabase_name(self, ctx):
        pass

    # Exit a parse tree produced by SQLiteParser#database_name.
    def exitDatabase_name(self, ctx):
        pass

    # Enter a parse tree produced by SQLiteParser#table_name.
    def enterTable_name(self, ctx: SQLiteParser.Table_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#table_name.
    def exitTable_name(self, ctx: SQLiteParser.Table_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#table_or_index_name.
    def enterTable_or_index_name(self, ctx: SQLiteParser.Table_or_index_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#table_or_index_name.
    def exitTable_or_index_name(self, ctx: SQLiteParser.Table_or_index_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#new_table_name.
    def enterNew_table_name(self, ctx: SQLiteParser.New_table_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#new_table_name.
    def exitNew_table_name(self, ctx: SQLiteParser.New_table_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#column_name.
    def enterColumn_name(self, ctx: SQLiteParser.Column_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#column_name.
    def exitColumn_name(self, ctx: SQLiteParser.Column_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#collation_name.
    def enterCollation_name(self, ctx: SQLiteParser.Collation_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#collation_name.
    def exitCollation_name(self, ctx: SQLiteParser.Collation_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#foreign_table.
    def enterForeign_table(self, ctx: SQLiteParser.Foreign_tableContext):
        pass

    # Exit a parse tree produced by SQLiteParser#foreign_table.
    def exitForeign_table(self, ctx: SQLiteParser.Foreign_tableContext):
        pass

    # Enter a parse tree produced by SQLiteParser#index_name.
    def enterIndex_name(self, ctx: SQLiteParser.Index_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#index_name.
    def exitIndex_name(self, ctx: SQLiteParser.Index_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#trigger_name.
    def enterTrigger_name(self, ctx: SQLiteParser.Trigger_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#trigger_name.
    def exitTrigger_name(self, ctx: SQLiteParser.Trigger_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#view_name.
    def enterView_name(self, ctx: SQLiteParser.View_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#view_name.
    def exitView_name(self, ctx: SQLiteParser.View_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#module_name.
    def enterModule_name(self, ctx: SQLiteParser.Module_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#module_name.
    def exitModule_name(self, ctx: SQLiteParser.Module_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#pragma_name.
    def enterPragma_name(self, ctx: SQLiteParser.Pragma_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#pragma_name.
    def exitPragma_name(self, ctx: SQLiteParser.Pragma_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#savepoint_name.
    def enterSavepoint_name(self, ctx: SQLiteParser.Savepoint_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#savepoint_name.
    def exitSavepoint_name(self, ctx: SQLiteParser.Savepoint_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#table_alias.
    def enterTable_alias(self, ctx: SQLiteParser.Table_aliasContext):
        pass

    # Exit a parse tree produced by SQLiteParser#table_alias.
    def exitTable_alias(self, ctx: SQLiteParser.Table_aliasContext):
        pass

    # Enter a parse tree produced by SQLiteParser#transaction_name.
    def enterTransaction_name(self, ctx: SQLiteParser.Transaction_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#transaction_name.
    def exitTransaction_name(self, ctx: SQLiteParser.Transaction_nameContext):
        pass

    # Enter a parse tree produced by SQLiteParser#any_name.
    def enterAny_name(self, ctx: SQLiteParser.Any_nameContext):
        pass

    # Exit a parse tree produced by SQLiteParser#any_name.
    def exitAny_name(self, ctx: SQLiteParser.Any_nameContext):
        pass


del SQLiteParser



"""
@file
@brief Convert R into Python
"""
from antlr4 import ParseTreeWalker
from pyquickhelper.pycode import remove_extra_spaces_and_pep8
from .RParser import RParser
from .RLexer import RLexer
from .antlr_grammar_use import parse_code
from .rconverterListener import TreeStringListener


def r2python(code: str, pep8=False, fLOG=None) -> str:
    """
    Converts a R script into Python.

    :param code: R string
    :param pep8: modify the output to be compliant with pep8
    :param fLOG: logging function
    :return: Python string

    .. _code-r2python:

    The function uses a customized R grammar implemented with Antlr4.
    Formula becomes strings. They should be handled with
    `patsy <http://patsy.readthedocs.io/en/latest/>`_.

    .. exref::
        :title: Convert R into Python

        .. runpython::
            :showcode:

            rscript = '''
                nb=function(y=1930){
                debut=1816
                MatDFemale=matrix(D$Female,nrow=111)
                colnames(MatDFemale)=(debut+0):198
                cly=(y-debut+1):111
                deces=diag(MatDFemale[:,cly[cly%in%1:199]])
                return(c(B$Female[B$Year==y],deces))}
                '''

            from pyensae.languages import r2python
            print(r2python(rscript, pep8=True))

    Some patterns are not well migrated such expression ``a:b`` into ``range(a,b)``.
    The grammar could be improved to detect the beginning of the expression but
    for now, if the function fails to do the conversion, ``a:b`` must be written
    into ``(a):b``. The same trick is sometimes needed for other patterns
    such as the operator ``%in%`` which is converted into an intersection
    of two sets.

    Kwonws bugs:

    * ``} else {`` needs to be replaced by ``} EOL else {``
    * comment added at the end of line must be followed by an empty line
    * ``m[,1]`` must be replaced by ``M[:,1]``
    * formula ``~.`` is not translated
    * ``%<%`` cannot be followed by an empty line

    The grammar were updated in 2022 for python 3.10 and
    :epkg:`antlr4-python3-runtime` == 4.10.
    """
    if fLOG:
        fLOG(  # pragma: no cover
            "[r2python] parse ", len(code), "bytes")
    parser = parse_code(code, RParser, RLexer)
    if fLOG:
        fLOG(  # pragma: no cover
            "[r2python] parse continue")
    parsed = parser.parse()
    if fLOG:
        fLOG(  # pragma: no cover
            "[r2python] listen")
    listen = TreeStringListener(parsed, fLOG=fLOG)
    walker = ParseTreeWalker()
    if fLOG:
        fLOG(  # pragma: no cover
            "[r2python] walk")
    walker.walk(listen, parsed)
    if fLOG:
        fLOG(  # pragma: no cover
            "[r2python] get code")
    code = listen.get_python()
    if pep8:
        if fLOG:
            fLOG(  # pragma: no cover
                "[r2python] pep8")
        code = remove_extra_spaces_and_pep8(code, aggressive=True)
    return code



# -*- coding: utf-8 -*-
"""
@file
@brief Magic commands about graphs

.. versionadded:: 1.1
"""

from IPython.core.magic import magics_class, line_magic
from pyquickhelper.ipythonhelper import MagicCommandParser, MagicClassWithHelpers

# do not import a module which imports matplotlib
# if this module is imported, this class is being tested and it affects sphinx
# when it generates the documentation
from .matplotlib_helper import mpl_switch_style


@magics_class
class MagicGraph(MagicClassWithHelpers):

    """
    Defines magic commands about graphs

    .. versionadded:: 1.1
    """

    @staticmethod
    def mpl_style_parser():
        """
        defines the way to parse the magic command ``%mpl_style``
        """
        parser = MagicCommandParser(
            description='changes matplotlib style', prog="mpl_style")
        parser.add_argument(
            'style',
            type=str,
            help='style, ggplot for exemple',
            default="ggplot")
        return parser

    @line_magic
    def mpl_style(self, line):
        """
        defines ``%mpl_style``
        which changes the style of matplotlib graphs, example: ``%mpl_style ggplot``

        .. nbref::
            :title: mpl_style

            This magic just does::

                import matplotlib.pyplot as plt
                plt.style.use('ggplot')

            It should take place at the beginning of the notebook.
        """
        parser = self.get_parser(MagicGraph.mpl_style_parser, "mpl_style")
        args = self.get_args(line, parser)

        if args is not None:
            style = args.style
            mpl_switch_style(style)


def register_graph_magics(ip=None):  # pragma: no cover
    """
    register magics function, can be called from a notebook

    @param      ip      from ``get_ipython()``
    """
    if ip is None:
        from IPython import get_ipython
        ip = get_ipython()
    ip.register_magics(MagicGraph)



# Generated from \R.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .RParser import RParser
else:
    from RParser import RParser

# This class defines a complete listener for a parse tree produced by RParser.


class RListener(ParseTreeListener):

    # Enter a parse tree produced by RParser#parse.
    def enterParse(self, ctx: RParser.ParseContext):
        pass

    # Exit a parse tree produced by RParser#parse.
    def exitParse(self, ctx: RParser.ParseContext):
        pass

    # Enter a parse tree produced by RParser#expr.
    def enterExpr(self, ctx: RParser.ExprContext):
        pass

    # Exit a parse tree produced by RParser#expr.
    def exitExpr(self, ctx: RParser.ExprContext):
        pass

    # Enter a parse tree produced by RParser#functiondefbody.
    def enterFunctiondefbody(self, ctx: RParser.FunctiondefbodyContext):
        pass

    # Exit a parse tree produced by RParser#functiondefbody.
    def exitFunctiondefbody(self, ctx: RParser.FunctiondefbodyContext):
        pass

    # Enter a parse tree produced by RParser#functiondeflambda.
    def enterFunctiondeflambda(self, ctx: RParser.FunctiondeflambdaContext):
        pass

    # Exit a parse tree produced by RParser#functiondeflambda.
    def exitFunctiondeflambda(self, ctx: RParser.FunctiondeflambdaContext):
        pass

    # Enter a parse tree produced by RParser#functiondefargslambda.
    def enterFunctiondefargslambda(self, ctx: RParser.FunctiondefargslambdaContext):
        pass

    # Exit a parse tree produced by RParser#functiondefargslambda.
    def exitFunctiondefargslambda(self, ctx: RParser.FunctiondefargslambdaContext):
        pass

    # Enter a parse tree produced by RParser#functiondefargs.
    def enterFunctiondefargs(self, ctx: RParser.FunctiondefargsContext):
        pass

    # Exit a parse tree produced by RParser#functiondefargs.
    def exitFunctiondefargs(self, ctx: RParser.FunctiondefargsContext):
        pass

    # Enter a parse tree produced by RParser#implicit_column_name.
    def enterImplicit_column_name(self, ctx: RParser.Implicit_column_nameContext):
        pass

    # Exit a parse tree produced by RParser#implicit_column_name.
    def exitImplicit_column_name(self, ctx: RParser.Implicit_column_nameContext):
        pass

    # Enter a parse tree produced by RParser#affectation.
    def enterAffectation(self, ctx: RParser.AffectationContext):
        pass

    # Exit a parse tree produced by RParser#affectation.
    def exitAffectation(self, ctx: RParser.AffectationContext):
        pass

    # Enter a parse tree produced by RParser#rangeopexpr.
    def enterRangeopexpr(self, ctx: RParser.RangeopexprContext):
        pass

    # Exit a parse tree produced by RParser#rangeopexpr.
    def exitRangeopexpr(self, ctx: RParser.RangeopexprContext):
        pass

    # Enter a parse tree produced by RParser#exprlist.
    def enterExprlist(self, ctx: RParser.ExprlistContext):
        pass

    # Exit a parse tree produced by RParser#exprlist.
    def exitExprlist(self, ctx: RParser.ExprlistContext):
        pass

    # Enter a parse tree produced by RParser#rightexpr.
    def enterRightexpr(self, ctx: RParser.RightexprContext):
        pass

    # Exit a parse tree produced by RParser#rightexpr.
    def exitRightexpr(self, ctx: RParser.RightexprContext):
        pass

    # Enter a parse tree produced by RParser#formlist.
    def enterFormlist(self, ctx: RParser.FormlistContext):
        pass

    # Exit a parse tree produced by RParser#formlist.
    def exitFormlist(self, ctx: RParser.FormlistContext):
        pass

    # Enter a parse tree produced by RParser#form.
    def enterForm(self, ctx: RParser.FormContext):
        pass

    # Exit a parse tree produced by RParser#form.
    def exitForm(self, ctx: RParser.FormContext):
        pass

    # Enter a parse tree produced by RParser#argumentname.
    def enterArgumentname(self, ctx: RParser.ArgumentnameContext):
        pass

    # Exit a parse tree produced by RParser#argumentname.
    def exitArgumentname(self, ctx: RParser.ArgumentnameContext):
        pass

    # Enter a parse tree produced by RParser#sublist.
    def enterSublist(self, ctx: RParser.SublistContext):
        pass

    # Exit a parse tree produced by RParser#sublist.
    def exitSublist(self, ctx: RParser.SublistContext):
        pass

    # Enter a parse tree produced by RParser#sublistadd.
    def enterSublistadd(self, ctx: RParser.SublistaddContext):
        pass

    # Exit a parse tree produced by RParser#sublistadd.
    def exitSublistadd(self, ctx: RParser.SublistaddContext):
        pass

    # Enter a parse tree produced by RParser#sub.
    def enterSub(self, ctx: RParser.SubContext):
        pass

    # Exit a parse tree produced by RParser#sub.
    def exitSub(self, ctx: RParser.SubContext):
        pass

    # Enter a parse tree produced by RParser#subnobracket.
    def enterSubnobracket(self, ctx: RParser.SubnobracketContext):
        pass

    # Exit a parse tree produced by RParser#subnobracket.
    def exitSubnobracket(self, ctx: RParser.SubnobracketContext):
        pass

    # Enter a parse tree produced by RParser#ranges.
    def enterRanges(self, ctx: RParser.RangesContext):
        pass

    # Exit a parse tree produced by RParser#ranges.
    def exitRanges(self, ctx: RParser.RangesContext):
        pass

    # Enter a parse tree produced by RParser#range_simple.
    def enterRange_simple(self, ctx: RParser.Range_simpleContext):
        pass

    # Exit a parse tree produced by RParser#range_simple.
    def exitRange_simple(self, ctx: RParser.Range_simpleContext):
        pass

    # Enter a parse tree produced by RParser#range_complexe.
    def enterRange_complexe(self, ctx: RParser.Range_complexeContext):
        pass

    # Exit a parse tree produced by RParser#range_complexe.
    def exitRange_complexe(self, ctx: RParser.Range_complexeContext):
        pass

    # Enter a parse tree produced by RParser#intersections.
    def enterIntersections(self, ctx: RParser.IntersectionsContext):
        pass

    # Exit a parse tree produced by RParser#intersections.
    def exitIntersections(self, ctx: RParser.IntersectionsContext):
        pass

    # Enter a parse tree produced by RParser#intersection_simple.
    def enterIntersection_simple(self, ctx: RParser.Intersection_simpleContext):
        pass

    # Exit a parse tree produced by RParser#intersection_simple.
    def exitIntersection_simple(self, ctx: RParser.Intersection_simpleContext):
        pass

    # Enter a parse tree produced by RParser#intersection_complexe.
    def enterIntersection_complexe(self, ctx: RParser.Intersection_complexeContext):
        pass

    # Exit a parse tree produced by RParser#intersection_complexe.
    def exitIntersection_complexe(self, ctx: RParser.Intersection_complexeContext):
        pass

    # Enter a parse tree produced by RParser#constant.
    def enterConstant(self, ctx: RParser.ConstantContext):
        pass

    # Exit a parse tree produced by RParser#constant.
    def exitConstant(self, ctx: RParser.ConstantContext):
        pass

    # Enter a parse tree produced by RParser#boolean.
    def enterBoolean(self, ctx: RParser.BooleanContext):
        pass

    # Exit a parse tree produced by RParser#boolean.
    def exitBoolean(self, ctx: RParser.BooleanContext):
        pass

    # Enter a parse tree produced by RParser#nextexpr.
    def enterNextexpr(self, ctx: RParser.NextexprContext):
        pass

    # Exit a parse tree produced by RParser#nextexpr.
    def exitNextexpr(self, ctx: RParser.NextexprContext):
        pass

    # Enter a parse tree produced by RParser#repeatexpr.
    def enterRepeatexpr(self, ctx: RParser.RepeatexprContext):
        pass

    # Exit a parse tree produced by RParser#repeatexpr.
    def exitRepeatexpr(self, ctx: RParser.RepeatexprContext):
        pass

    # Enter a parse tree produced by RParser#whileexpr.
    def enterWhileexpr(self, ctx: RParser.WhileexprContext):
        pass

    # Exit a parse tree produced by RParser#whileexpr.
    def exitWhileexpr(self, ctx: RParser.WhileexprContext):
        pass

    # Enter a parse tree produced by RParser#forexpr.
    def enterForexpr(self, ctx: RParser.ForexprContext):
        pass

    # Exit a parse tree produced by RParser#forexpr.
    def exitForexpr(self, ctx: RParser.ForexprContext):
        pass

    # Enter a parse tree produced by RParser#ifexpr.
    def enterIfexpr(self, ctx: RParser.IfexprContext):
        pass

    # Exit a parse tree produced by RParser#ifexpr.
    def exitIfexpr(self, ctx: RParser.IfexprContext):
        pass

    # Enter a parse tree produced by RParser#ifelseexpr.
    def enterIfelseexpr(self, ctx: RParser.IfelseexprContext):
        pass

    # Exit a parse tree produced by RParser#ifelseexpr.
    def exitIfelseexpr(self, ctx: RParser.IfelseexprContext):
        pass

    # Enter a parse tree produced by RParser#elseif.
    def enterElseif(self, ctx: RParser.ElseifContext):
        pass

    # Exit a parse tree produced by RParser#elseif.
    def exitElseif(self, ctx: RParser.ElseifContext):
        pass

    # Enter a parse tree produced by RParser#returnexpr.
    def enterReturnexpr(self, ctx: RParser.ReturnexprContext):
        pass

    # Exit a parse tree produced by RParser#returnexpr.
    def exitReturnexpr(self, ctx: RParser.ReturnexprContext):
        pass

    # Enter a parse tree produced by RParser#functioncall.
    def enterFunctioncall(self, ctx: RParser.FunctioncallContext):
        pass

    # Exit a parse tree produced by RParser#functioncall.
    def exitFunctioncall(self, ctx: RParser.FunctioncallContext):
        pass

    # Enter a parse tree produced by RParser#inlinefunction.
    def enterInlinefunction(self, ctx: RParser.InlinefunctionContext):
        pass

    # Exit a parse tree produced by RParser#inlinefunction.
    def exitInlinefunction(self, ctx: RParser.InlinefunctionContext):
        pass

    # Enter a parse tree produced by RParser#formula_simple.
    def enterFormula_simple(self, ctx: RParser.Formula_simpleContext):
        pass

    # Exit a parse tree produced by RParser#formula_simple.
    def exitFormula_simple(self, ctx: RParser.Formula_simpleContext):
        pass

    # Enter a parse tree produced by RParser#formula_simple_A.
    def enterFormula_simple_A(self, ctx: RParser.Formula_simple_AContext):
        pass

    # Exit a parse tree produced by RParser#formula_simple_A.
    def exitFormula_simple_A(self, ctx: RParser.Formula_simple_AContext):
        pass

    # Enter a parse tree produced by RParser#formula_simple_B.
    def enterFormula_simple_B(self, ctx: RParser.Formula_simple_BContext):
        pass

    # Exit a parse tree produced by RParser#formula_simple_B.
    def exitFormula_simple_B(self, ctx: RParser.Formula_simple_BContext):
        pass

    # Enter a parse tree produced by RParser#formula_simple_C.
    def enterFormula_simple_C(self, ctx: RParser.Formula_simple_CContext):
        pass

    # Exit a parse tree produced by RParser#formula_simple_C.
    def exitFormula_simple_C(self, ctx: RParser.Formula_simple_CContext):
        pass

    # Enter a parse tree produced by RParser#affectop.
    def enterAffectop(self, ctx: RParser.AffectopContext):
        pass

    # Exit a parse tree produced by RParser#affectop.
    def exitAffectop(self, ctx: RParser.AffectopContext):
        pass

    # Enter a parse tree produced by RParser#functiondef.
    def enterFunctiondef(self, ctx: RParser.FunctiondefContext):
        pass

    # Exit a parse tree produced by RParser#functiondef.
    def exitFunctiondef(self, ctx: RParser.FunctiondefContext):
        pass

    # Enter a parse tree produced by RParser#identifier.
    def enterIdentifier(self, ctx: RParser.IdentifierContext):
        pass

    # Exit a parse tree produced by RParser#identifier.
    def exitIdentifier(self, ctx: RParser.IdentifierContext):
        pass

    # Enter a parse tree produced by RParser#formop.
    def enterFormop(self, ctx: RParser.FormopContext):
        pass

    # Exit a parse tree produced by RParser#formop.
    def exitFormop(self, ctx: RParser.FormopContext):
        pass

    # Enter a parse tree produced by RParser#rangeop.
    def enterRangeop(self, ctx: RParser.RangeopContext):
        pass

    # Exit a parse tree produced by RParser#rangeop.
    def exitRangeop(self, ctx: RParser.RangeopContext):
        pass

    # Enter a parse tree produced by RParser#dotop.
    def enterDotop(self, ctx: RParser.DotopContext):
        pass

    # Exit a parse tree produced by RParser#dotop.
    def exitDotop(self, ctx: RParser.DotopContext):
        pass

    # Enter a parse tree produced by RParser#operator.
    def enterOperator(self, ctx: RParser.OperatorContext):
        pass

    # Exit a parse tree produced by RParser#operator.
    def exitOperator(self, ctx: RParser.OperatorContext):
        pass

    # Enter a parse tree produced by RParser#comparison.
    def enterComparison(self, ctx: RParser.ComparisonContext):
        pass

    # Exit a parse tree produced by RParser#comparison.
    def exitComparison(self, ctx: RParser.ComparisonContext):
        pass


del RParser



# -*- coding: utf-8 -*-
"""
@file
@brief Parses :epkg:`C# `.
"""
from antlr4.tree.Tree import TerminalNodeImpl
from antlr4 import ParseTreeWalker
from ..languages.antlr_grammar_use import get_parser_lexer, parse_code
from ..languages.CSharpParserListener import CSharpParserListener


class CSharpElement:
    """
    Base class of a :epkg:`C#` element.
    """

    _kinds = {'class', 'method', 'domain', 'type'}
    _privates = {'public', 'private', 'protected'}

    def __init__(self, domain, name, doc=None, source=None, code=None, **kwargs):
        """
        @param  domain      domain
        @param  name        name
        @param  doc         documentation
        @param  source      source file
        @param  code        code
        """
        self.name = name
        self.doc = doc
        self.domain = domain
        self.source = source
        self.code = code
        self.options = kwargs

    def __str__(self):
        """
        usual
        """
        return f"{self.name}"


class CSharpDomain(CSharpElement):
    """
    Base class of a :epkg:`C#` domaon.
    """

    def __init__(self, domain, name, doc=None, source=None, code=None, **kwargs):
        """
        @param  domain      domain
        @param  name        name
        @param  doc         documentation
        @param  source      source file
        @param  code        code
        """
        if "{" in domain or "{" in name:
            raise ValueError(
                "Issue with domain name\ndomain={0}\nname={1}".format(domain, name))
        CSharpElement.__init__(self, domain, name, doc, source, code, **kwargs)

    def __str__(self):
        """
        usual
        """
        return f"namespace {self.name}"


class CSharpVariable(CSharpElement):
    """
    :epkg:`C#` variables
    """

    def __init__(self, domain, name, typ, value=None, doc=None, source=None, code=None, **kwargs):
        """
        @param      domain      domain
        @param      name        name
        @param      typ         type
        @param      value       default value if it exists
        @param      doc         documentation
        @param  source      source file
        @param  code        code
        """
        CSharpElement.__init__(self, domain, name, doc=doc,
                               source=source, code=code, **kwargs)
        if not isinstance(typ, CSharpType):
            raise TypeError("typ must be of type CSharpType")
        self.type = typ
        self.value = value

    def __str__(self):
        """
        usual
        """
        if isinstance(self.value, str):
            v = self.value.replace('"', '\\"')
            value = f'"{v}"'
        else:
            value = self.value
        return f"{self.type} {self.name} = {value}"


class CSharpType(CSharpVariable):
    """
    :epkg:`C#` type
    """

    def __init__(self, domain, name, typ, doc=None, source=None, code=None, **kwargs):
        """
        @param      domain      domain
        @param      name        name
        @param      typ         type
        @param      source      source file
        @param      code        code
        """
        CSharpVariable.__init__(
            self, domain, name, typ=typ, doc=doc, source=source, code=code, **kwargs)

    def __str__(self):
        """
        usual
        """
        return f"{self.name}"


class CSharpMethod(CSharpElement):
    """
    :epkg:`C#` function.
    """

    def __init__(self, domain, name, rtype, params=None, private='public',
                 static=False, doc=None, source=None, code=None, **kwargs):
        """
        @param      domain          domain
        @param      name            name
        @param      rtype           return type
        @param      private         private, public or protected
        @param      static          static or not
        @param      parameters      parameters
        @param      doc             documentation
        @param      code            code
        """
        CSharpElement.__init__(self, domain, name, doc=doc,
                               source=source, code=code, **kwargs)
        if private not in CSharpElement._privates:
            raise ValueError(
                f"Unable to find '{private}' in {CSharpElement._privates}.")
        self.params = params
        self.source = source
        self.static = static
        self.private = private
        self.rtype = rtype

    def __str__(self):
        """
        usual
        """
        pars = ", ".join(str(p) for p in self.params)
        return f"{self.domain}.{self.name} ({self.kind})"


class CSharpClass(CSharpElement):
    """
    :epkg:`C#` class.
    """

    def __init__(self, domain, name, methods=None, constants=None,
                 params=None, private='public', static=False,
                 doc=None, source=None, code=None, **kwargs):
        """
        @param      domain          domain
        @param      name            name
        @param      methods         methods
        @param      constants       constants
        @param      private         private, public or protected
        @param      static          static or not
        @param      doc             documentation
        @param      source          source file
        @param      code            code
        """
        CSharpElement.__init__(self, domain, name, doc=doc,
                               source=source, code=code, **kwargs)
        if private not in CSharpElement._privates:
            raise ValueError(
                f"Unable to find '{private}' in {CSharpElement._privates}.")
        self.source = source
        self.static = static
        self.private = private
        self.methods = methods
        self.contansts = constants

    def __str__(self):
        """
        usual
        """
        return f"class {self.domain}.{self.name}"


class CSharpParserListenerSignatures(CSharpParserListener):
    """
    :epkg:`C#` Listener
    """

    def __init__(self, parser, source):
        """
        constructor

        @param      parser      parser used to parse the code
        """
        CSharpParserListener.__init__(self)
        self.buffer = []
        self.parser = parser
        self._elements = []
        self._context = []
        self._obj = None
        self._source = source

    def stack_element(self, el):
        """
        Adds an element.
        """
        self._obj = el

    def enter_body(self):
        """
        Adds an element.
        """
        if self._obj is None:
            raise RuntimeError("self._obj cannot be None")
        self._context.append(self._obj)

    def exit_body(self):
        """
        Adds an element.
        """
        el = self._context.pop()
        if el is None:
            raise RuntimeError("el cannot be None")
        self._elements.append(el)

    # def enterEveryRule(self, ctx):
    #    pass

    # def leaveEveryRule(self, ctx):
    #    pass

    def get_code(self, ctx):
        line, col = ctx.start.line, ctx.start.column
        class_name = ctx.__class__.__name__
        return class_name, line, col

    def enumerate_all_children(self, ctx):
        """
        Enumerate all children.

        @param      ctx     context
        @return             iterator
        """

        def roll(ctx):
            if not isinstance(ctx, TerminalNodeImpl):
                yield ctx
                for ch in ctx.getChildren():
                    for r in roll(ch):
                        yield r
        for r in roll(ctx):
            yield r

    ############
    # namespace
    ############

    def enterNamespace_declaration(self, ctx):
        self._namespace_code = ctx.getText()

    def enterNamespace_body(self, ctx):
        pass

    def enterNamespace_or_type_name(self, ctx):
        code = getattr(self, '_namespace_code', None)
        if code:
            self._namespace_code = None
            name = ctx.getText()
            children = list(self.enumerate_all_children(ctx))
            line1 = ctx.start.line
            line2 = ctx.stop.line
            self.stack_element(CSharpDomain(
                name, name, code=code, source=self._source, lines=[line1, line2]))
            self.enter_body()

    def exitNamespace_or_type_name(self, ctx):
        pass

    def exitNamespace_body(self, ctx):
        self.exit_body()

    def exitNamespace_declaration(self, ctx):
        pass


class CSharpParser:
    """
    Parses :epkg:`C#`.
    """

    def __init__(self):
        """
        """
        clparser, cllexer = get_parser_lexer("C#")
        self._parser = clparser
        self._lexer = cllexer

    def parse(self, code, source=None):
        """
        Returns all elements of codes inside a string.

        @param      code        string
        @param      source      source
        @return                 list of @see cl CSharpElement
        """
        self._source = source
        parser = parse_code(code, self._parser, self._lexer)
        tree = parser.compilation_unit()
        walker = ParseTreeWalker()
        listen = CSharpParserListenerSignatures(parser, source)
        walker.walk(listen, tree)
        return listen._elements



# Generated from \CSharpPreprocessorParser.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .CSharpPreprocessorParser import CSharpPreprocessorParser
else:
    from CSharpPreprocessorParser import CSharpPreprocessorParser
import java.util.Stack
import java.util.HashSet

# This class defines a complete listener for a parse tree produced by CSharpPreprocessorParser.


class CSharpPreprocessorParserListener(ParseTreeListener):

    # Enter a parse tree produced by CSharpPreprocessorParser#preprocessorDeclaration.
    def enterPreprocessorDeclaration(self, ctx: CSharpPreprocessorParser.PreprocessorDeclarationContext):
        pass

    # Exit a parse tree produced by CSharpPreprocessorParser#preprocessorDeclaration.
    def exitPreprocessorDeclaration(self, ctx: CSharpPreprocessorParser.PreprocessorDeclarationContext):
        pass

    # Enter a parse tree produced by CSharpPreprocessorParser#preprocessorConditional.
    def enterPreprocessorConditional(self, ctx: CSharpPreprocessorParser.PreprocessorConditionalContext):
        pass

    # Exit a parse tree produced by CSharpPreprocessorParser#preprocessorConditional.
    def exitPreprocessorConditional(self, ctx: CSharpPreprocessorParser.PreprocessorConditionalContext):
        pass

    # Enter a parse tree produced by CSharpPreprocessorParser#preprocessorLine.
    def enterPreprocessorLine(self, ctx: CSharpPreprocessorParser.PreprocessorLineContext):
        pass

    # Exit a parse tree produced by CSharpPreprocessorParser#preprocessorLine.
    def exitPreprocessorLine(self, ctx: CSharpPreprocessorParser.PreprocessorLineContext):
        pass

    # Enter a parse tree produced by CSharpPreprocessorParser#preprocessorDiagnostic.
    def enterPreprocessorDiagnostic(self, ctx: CSharpPreprocessorParser.PreprocessorDiagnosticContext):
        pass

    # Exit a parse tree produced by CSharpPreprocessorParser#preprocessorDiagnostic.
    def exitPreprocessorDiagnostic(self, ctx: CSharpPreprocessorParser.PreprocessorDiagnosticContext):
        pass

    # Enter a parse tree produced by CSharpPreprocessorParser#preprocessorRegion.
    def enterPreprocessorRegion(self, ctx: CSharpPreprocessorParser.PreprocessorRegionContext):
        pass

    # Exit a parse tree produced by CSharpPreprocessorParser#preprocessorRegion.
    def exitPreprocessorRegion(self, ctx: CSharpPreprocessorParser.PreprocessorRegionContext):
        pass

    # Enter a parse tree produced by CSharpPreprocessorParser#preprocessorPragma.
    def enterPreprocessorPragma(self, ctx: CSharpPreprocessorParser.PreprocessorPragmaContext):
        pass

    # Exit a parse tree produced by CSharpPreprocessorParser#preprocessorPragma.
    def exitPreprocessorPragma(self, ctx: CSharpPreprocessorParser.PreprocessorPragmaContext):
        pass

    # Enter a parse tree produced by CSharpPreprocessorParser#directive_new_line_or_sharp.
    def enterDirective_new_line_or_sharp(self, ctx: CSharpPreprocessorParser.Directive_new_line_or_sharpContext):
        pass

    # Exit a parse tree produced by CSharpPreprocessorParser#directive_new_line_or_sharp.
    def exitDirective_new_line_or_sharp(self, ctx: CSharpPreprocessorParser.Directive_new_line_or_sharpContext):
        pass

    # Enter a parse tree produced by CSharpPreprocessorParser#preprocessor_expression.
    def enterPreprocessor_expression(self, ctx: CSharpPreprocessorParser.Preprocessor_expressionContext):
        pass

    # Exit a parse tree produced by CSharpPreprocessorParser#preprocessor_expression.
    def exitPreprocessor_expression(self, ctx: CSharpPreprocessorParser.Preprocessor_expressionContext):
        pass


del CSharpPreprocessorParser



"""
@file
@brief CorrPlot functionalities.

It comes from `corrplot.py <https://raw.githubusercontent.com/biokit/biokit/master/biokit/viz/corrplot.py>`_
which I copied here because the module does not properly work on Python 3 (import issues).
See also `biokit license <https://github.com/biokit/biokit/blob/master/LICENSE>`_.

:author: Thomas Cokelaer
:references: http://cran.r-project.org/web/packages/corrplot/vignettes/corrplot-intro.html
"""
import numpy as np


class Color:
    """
    Lists of known colors.
    """

    colors = {
        'red': (1., 0., 0.),
        'green': (0., 1., 0.),
        'blue': (0., 0., 1.),
        'white': (1., 1., 1.),
        'black': (0., 0., 0.),
        'darkblue': '#00008b',
    }

    def __init__(self, name):
        """
        @param  name    hexadecimal or name
        """
        if len(name) == 7 and name[0] == '#':
            self.set_hex(name)
        else:
            value = Color.colors.get(name, None)
            if value is None:
                raise ValueError("Unknown color name '{}'.".format(name))
            if len(value) == 7:
                self.set_hex(value)
            else:
                self.red, self.green, self.blue = value

    def set_hex(self, name):
        """
        Converts a string like ``#AABBCC`` into `(red, green, blue)`.
        """
        self.red = float.fromhex('0x' + name[1:3].lower()) / 255
        self.green = float.fromhex('0x' + name[3:5].lower()) / 255
        self.blue = float.fromhex('0x' + name[5:7].lower()) / 255


class Colormap:
    """
    Snippets of code from `colormap/colors.py
    <https://github.com/cokelaer/colormap/blob/master/src/colormap/colors.py>`_,
    `LICENSE <https://github.com/cokelaer/colormap/blob/master/LICENSE>`_.
    """
    # Source:

    def _get_colormap_mpl(self):
        try:
            from matplotlib.pyplot import colormaps as _cmaps
            return _cmaps()
        except ImportError:  # pragma: no cover
            return []
    colormaps = property(_get_colormap_mpl)

    def _get_diverging_black(self):
        return ['red_black_sky', 'red_black_blue', 'red_black_green', 'yellow_black_blue',
                'yellow_black_sky', 'red_black_orange', 'pink_black_green(w3c)']
    diverging_black = property(_get_diverging_black)

    def cmap_linear(self, color1, color2, color3, reverse=False, N=256):
        """
        Provides 3 colors in format accepted by :class:`Color`
        ::
            red = Color('red')
            cmap = cmap_linear(red, 'white', '#0000FF')
        """
        c1 = Color(color1)
        c2 = Color(color2)
        c3 = Color(color3)
        dico = {'red': [c1.red, c2.red, c3.red],
                'green': [c1.green, c2.green, c3.green],
                'blue': [c1.blue, c2.blue, c3.blue]}

        return self.cmap(dico, reverse=reverse, N=N)

    def cmap(self, colors=None, reverse=False, N=256):
        """
        Returns a colormap object to be used within matplotlib
        :param dict colors: a dictionary that defines the RGB colors to be
            used in the colormap. See :meth:`get_cmap_heat` for an example.
        :param bool reverse: reverse the colormap is  set to True (defaults to False)
        :param int N: Defaults to 50
        """
        # matplotlib colormaps
        if colors in self.colormaps:
            if reverse and colors.endswith("_r") is False:
                colors += "_r"
            from matplotlib.cm import get_cmap
            return get_cmap(colors)
        # custom ones
        if colors in self.diverging_black:
            c1, c2, c3 = colors.split("_")
            # special case of sky, which does not exists
            c3 = c3.replace("sky", "deep sky blue")
            return self.cmap_linear(c1, c2, c3)
        if colors == 'heat':
            return self.get_cmap_heat()
        if colors == 'heat_r':
            return self.get_cmap_heat_r()

        if reverse:
            for k in colors.keys():
                colors[k].reverse()

        # If index not given, RGB colors are evenly-spaced in colormap.
        index = np.linspace(0, 1, len(colors['red']))

        # Adapt color_data to the form expected by LinearSegmentedColormap.
        color_data = dict((key, [(x, y, y) for x, y in zip(index, value)])
                          for key, value in list(colors.items()))

        import matplotlib.colors
        m = matplotlib.colors.LinearSegmentedColormap(
            'my_color_map', color_data, N)
        return m


def cmap_builder(name, name2=None, name3=None):
    """
    Returns a colormap object compatible with matplotlib
    If only parameter **name** is provided, it should be a known matplotlib
    colormap name (e.g., jet). If **name2** is provided, then a new colormap
    is created going from the color **name** to the color **name2** with a
    linear scale. Finally, if **name3** is provided, a linear scaled colormap
    is built from color **name** to color **name3** with the intermediate color
    being the **name2**.
    Matplotlib colormap map names
    Source: `colormap/get_cmap.py
    <https://github.com/cokelaer/colormap/blob/master/src/colormap/get_cmap.py>`_,
    `LICENSE <https://github.com/cokelaer/colormap/blob/master/LICENSE>`_).
    """
    c = Colormap()
    # an R colormap
    if name and name2 and name3:
        return c.cmap_linear(name, name2, name3)
    if name and name2:
        return c.cmap_bicolor(name, name2)
    if name == 'heat':
        return c.get_cmap_heat()
    if name == 'heat_r':
        return c.get_cmap_heat_r()
    # matplotlic colormaps
    if name in c.colormaps:
        return c.cmap(name)
    # some custom diverging colormaps with black in the middle.
    if name in c.diverging_black:
        return c.cmap(name)
    if name.count("_") == 2:
        name1, name2, name3 = name.split("_")
        return c.cmap_linear(name1, name2, name3)

    # valid = c.colormaps + c.diverging_black
    txt = "name provided {0} is not recognised. ".format(name)
    txt += "\n valid name can be found in colormap.colormap_names"
    raise ValueError(txt)



"""
@file
@brief Various functions to decompress files
"""
import zipfile
import os
import gzip
import bz2
import warnings
import copy
import tarfile
from tarfile import ExtractError
from pyquickhelper.loghelper import noLOG


def decompress_zip(filename, whereTo=".", fLOG=noLOG):
    """
    Unzips a :epkg:`zip` file.

    @param      filename        file to process
    @param      whereTo         location of the result
    @param      fLOG            logging function
    @return                     return the list of decompressed files
    """
    try:
        file = zipfile.ZipFile(filename, "r")
    except zipfile.BadZipFile as e:  # pragma: no cover
        raise RuntimeError("Unable to unzip '{}'.".format(
            filename)) from e
    files = []
    for info in file.infolist():
        if not os.path.exists(info.filename):
            data = file.read(info.filename)
            tos = os.path.join(whereTo, info.filename)
            if not os.path.exists(tos):
                finalfolder = os.path.split(tos)[0]
                if not os.path.exists(finalfolder):
                    fLOG(  # pragma: no cover
                        "[decompress_zip] creating folder '{0}'".format(
                            finalfolder))
                    os.makedirs(finalfolder)  # pragma: no cover
                if not info.filename.endswith("/"):
                    u = open(tos, "wb")
                    u.write(data)
                    u.close()
                    files.append(tos)
                    fLOG("[decompress_zip] unzipped '{0}' to '{1}'".format(
                        info.filename, tos))
            elif not tos.endswith("/"):
                files.append(tos)
        elif not info.filename.endswith("/"):
            files.append(info.filename)
    return files


def extractall_silent(self, path=".", members=None, *, numeric_owner=False, silent=False):
    """
    Extracts all members from the archive to the current working
    directory and set owner, modification time and permissions on
    directories afterwards. `path' specifies a different directory
    to extract to. `members' is optional and must be a subset of the
    list returned by getmembers(). If `numeric_owner` is True, only
    the numbers for user/group names are used and not the names.

    Same function as `TarFile.extractall <https://github.com/python/cpython/blob/master/Lib/tarfile.py>`_
    but raises a warning if something wrong happens if silent is True.
    """
    directories = []

    if members is None:
        members = self

    for tarinfo in members:
        if tarinfo.isdir():
            # Extract directories with a safe mode.
            directories.append(tarinfo)
            tarinfo = copy.copy(tarinfo)
            tarinfo.mode = 0o700
        # Do not set_attrs directories, as we will do that further down
        if silent:  # pragma: no cover
            try:
                self.extract(tarinfo, path, set_attrs=not tarinfo.isdir(),
                             numeric_owner=numeric_owner)
            except FileNotFoundError as e:
                warnings.warn(
                    "[TarFile.extractall_silent] issue with '{0}' - {1}".format(path, e))
        else:
            self.extract(tarinfo, path, set_attrs=not tarinfo.isdir(),
                         numeric_owner=numeric_owner)

    # Reverse sort directories.
    directories.sort(key=lambda a: a.name)
    directories.reverse()

    # Set correct owner, mtime and filemode on directories.
    for tarinfo in directories:
        dirpath = os.path.join(path, tarinfo.name)
        try:
            self.chown(tarinfo, dirpath, numeric_owner=numeric_owner)
            self.utime(tarinfo, dirpath)
            self.chmod(tarinfo, dirpath)
        except ExtractError as e:  # pragma: no cover
            if self.errorlevel > 1:
                raise
            self._dbg(1, "tarfile: %s" % e)


def decompress_targz(filename, whereTo=".", silent=True, fLOG=noLOG):
    """
    Decompresses a :epkg:`tar.gz` file.

    @param      filename        file to process
    @param      folder          location of the result
    @param      silent          raise a warning instead of an error
    @param      fLOG            logging function
    @return                     return the list of decompressed files
    """
    tfile = tarfile.open(filename, 'r:gz')
    files = tfile.getmembers()
    extractall_silent(tfile, whereTo, silent=silent)
    t = [os.path.join(whereTo, f.name) for f in files]
    return [f for f in t if os.path.isfile(f)]


def decompress_gz(filename, whereTo=".", fLOG=noLOG):
    """
    Decompresses a :epkg:`tar.gz` file.

    @param      filename        file to process
    @param      folder          location of the result
    @param      fLOG            logging function
    @return                     return the list of decompressed files (only one)
    """
    if not filename.endswith(".gz"):
        raise NameError(  # pragma: no cover
            "the file should end with .gz: %r" % filename)
    dest = os.path.join(whereTo, filename[:-3])
    with gzip.open(filename, 'rb') as f:
        with open(dest, "wb") as g:
            g.write(f.read())
    return [dest]


def decompress_bz2(filename, whereTo=".", fLOG=noLOG):
    """
    Decompresses a :epkg:`bz2` file.

    @param      filename        file to process
    @param      folder          location of the result
    @param      fLOG            logging function
    @return                     return the list of decompressed files (only one)
    """
    if not filename.endswith(".bz2"):
        raise NameError(  # pragma: no cover
            "the file should end with .bz2 not '{0}'".format(filename))
    dest = os.path.join(whereTo, os.path.split(filename)[-1][:-4])
    with bz2.open(filename, 'rb') as f:
        with open(dest, "wb") as g:
            g.write(f.read())
    return [dest]
